METRIC_PREFIXES = {milli: 10**-3, centi: 10**-2, deci: 10**-1, '': 1, deka: 10**1, hecto: 10**2, kilo: 10**3}
IMPERIAL_UNITS = {inches: 12**-1, feet: 1, yards: 3, miles: 5280}
DIMENSIONS = {'': 1, square_: 2, cubic_: 3}
FEET_TO_METRES_CONVERSION = 0.3048

require 'active_support/inflector'

Numeric.class_eval do
  DIMENSIONS.each do |dimension_prefix, dimension|
    METRIC_PREFIXES.keys.each do|prefix|
      define_method("#{dimension_prefix}#{prefix}metres") do
        distance_ob = MetricDistance.new(self, prefix.intern, dimension, dimension_prefix)
        return distance_ob
      end
      alias_method "#{dimension_prefix}#{prefix}metre".intern, "#{dimension_prefix}#{prefix}metres".intern
    end
  end

  DIMENSIONS.each do |dimension_prefix, dimension|
    IMPERIAL_UNITS.keys.each do |base_unit|
      define_method("#{dimension_prefix}#{base_unit}") do
        distance_ob = ImperialDistance.new(self, base_unit, dimension, dimension_prefix)
        return distance_ob
      end
      alias_method "#{dimension_prefix}#{base_unit.to_s.singularize}".intern, "#{dimension_prefix}#{base_unit}".intern
    end
    alias_method "#{dimension_prefix}foot".intern, "#{dimension_prefix}feet".intern
  end

end

class DistanceUnit < Numeric

  attr_reader :dimension, :dimension_prefix
  attr_accessor :value, :base_unit

  def initialize value, base_unit, dimension, dimension_prefix
    @value = value
    @base_unit = base_unit
    @dimension = dimension
    @dimension_prefix = dimension_prefix
  end

  def format_display
    display_string = ''
    base_distance = self.to_metres
    if base_distance.value >= 1000
      display_string = "#{base_distance.to_kilometres.value.round(1)} km"
    else
      display_string = "#{base_distance.value.round(0)} m"
    end
    return display_string
  end

  def unit
    return "#{self.dimension_prefix}#{self.base_unit}"
  end

  def value
    return @value.to_f
  end

  DIMENSIONS.each do |dimension_prefix, dimension|
    METRIC_PREFIXES.each do |prefix, scale|
      define_method("to_#{dimension_prefix}#{prefix}metres") do
        raise "Dimension Error" if self.dimension != dimension
        base_imperial = self.send("to_#{dimension_prefix}feet")
        metric_value = base_imperial.value * (FEET_TO_METRES_CONVERSION ** dimension)
        base_metric = MetricDistance.new(metric_value, '', dimension, dimension_prefix)
        metric_distance = base_metric.send("to_#{dimension_prefix}#{prefix}metres")
      end
      alias_method "to_#{dimension_prefix}#{prefix}metre".intern, "to_#{dimension_prefix}#{prefix}metres".intern
    end
  end

  DIMENSIONS.each do |dimension_prefix, dimension|
    IMPERIAL_UNITS.each do |base_unit, scale|
      define_method("to_#{dimension_prefix}#{base_unit}") do
        raise "Dimension Error" if self.dimension != dimension
        base_metric = self.send("to_#{dimension_prefix}metres")
        imperial_value = base_metric.value * (FEET_TO_METRES_CONVERSION ** -dimension)
        base_imperial = ImperialDistance.new(imperial_value, :feet, dimension, dimension_prefix)
        imperial_distance = base_imperial.send("to_#{dimension_prefix}#{base_unit}")
      end
      alias_method "to_#{dimension_prefix}#{base_unit.to_s.singularize}".intern, "to_#{dimension_prefix}#{base_unit}".intern
    end
    alias_method "to_#{dimension_prefix}foot".intern, "to_#{dimension_prefix}feet".intern
  end

end

class MetricDistance < DistanceUnit

  attr_accessor :metric_prefix

  def initialize value, metric_prefix, dimension, dimension_prefix
    @metric_prefix = metric_prefix
    base_unit = "#{metric_prefix}metres".intern
    super(value, base_unit, dimension, dimension_prefix)
  end

  def metric_prefix= metric_prefix
    self.metric_prefix = metric_prefix
    self.base_unit = "#{metric_prefix}metres".intern
  end

  DIMENSIONS.each do |dimension_prefix, dimension|
    METRIC_PREFIXES.each do |prefix, scale|
      define_method("to_#{dimension_prefix}#{prefix}metres") do
        raise "Dimension Error" if self.dimension != dimension
        current_value = self.value
        current_prefix = self.metric_prefix
        new_value = current_value * ((METRIC_PREFIXES[current_prefix.intern] * (scale**-1)) ** dimension)
        return MetricDistance.new(new_value, prefix.intern, dimension, dimension_prefix)
      end
      alias_method "to_#{dimension_prefix}#{prefix}metre".intern, "to_#{dimension_prefix}#{prefix}metres".intern
    end
  end

end

class ImperialDistance < DistanceUnit

  DIMENSIONS.each do |dimension_prefix, dimension|
    IMPERIAL_UNITS.each do |base_unit, scale|
      define_method("to_#{dimension_prefix}#{base_unit}") do
        raise "Dimension Error" if self.dimension != dimension
        current_value = self.value
        current_unit = self.base_unit
        new_value = current_value * ((IMPERIAL_UNITS[current_unit.intern] * (scale**-1)) ** dimension)
        return ImperialDistance.new(new_value, base_unit, dimension, dimension_prefix)
      end
      alias_method "to_#{dimension_prefix}#{base_unit.to_s.singularize}".intern, "to_#{dimension_prefix}#{base_unit}".intern
    end
    alias_method "to_#{dimension_prefix}foot".intern, "to_#{dimension_prefix}feet".intern
  end

end


